import datetime
import logging
import time
from typing import List, Callable

from .. import SenseidReaderConnectionInfo, SupportedSenseidReader
from .multicast_dns_service_discovery import MulticastDnsServiceDiscoveryScanner
from .serialport import SerialPortScanner

logger = logging.getLogger(__name__)


class SenseidReaderScanner:

    def __init__(self, notification_callback: Callable[[SenseidReaderConnectionInfo], None] = None,
                 autostart: bool = False):
        logger.info('Starting Reader Scanner')
        self.notification_callback = notification_callback
        self.readers: List[SenseidReaderConnectionInfo] = []
        self.serial_port_scanner = SerialPortScanner(notification_callback=self._add_reader)
        self.multicast_dns_service_discovery_scanner = MulticastDnsServiceDiscoveryScanner(
            notification_callback=self._add_reader)
        if autostart:
            self.start()

    def start(self, reset: bool = False, notification_callback: Callable[[SenseidReaderConnectionInfo], None] = None):
        if reset:
            self.readers: List[SenseidReaderConnectionInfo] = []
        if notification_callback is not None:
            self.notification_callback = notification_callback

        self.serial_port_scanner.start(reset=reset)
        self.multicast_dns_service_discovery_scanner.start()

    def stop(self):
        self.serial_port_scanner.stop()
        self.multicast_dns_service_discovery_scanner.stop()

    def _add_reader(self, connection_info: SenseidReaderConnectionInfo):
        self.readers.append(connection_info)
        if self.notification_callback is not None:
            self.notification_callback(connection_info)

    def get_readers(self):
        return self.readers

    def _get_first_reader_connection_info_of_type(self, reader_type: SupportedSenseidReader):
        for reader_connection_info in self.readers:
            if reader_connection_info.driver == reader_type:
                return reader_connection_info

    def wait_for_reader_of_type(self, reader_type: SupportedSenseidReader, timeout_s=-1) -> SenseidReaderConnectionInfo:
        timestamp_timeout = (datetime.datetime.now() + datetime.timedelta(seconds=timeout_s)) if timeout_s > 0 else None
        while True:
            reader_connection_info = self._get_first_reader_connection_info_of_type(reader_type)
            if reader_connection_info is not None:
                return reader_connection_info
            time.sleep(0.1)
            if timestamp_timeout is not None:
                if datetime.datetime.now() > timestamp_timeout:
                    break



import logging
from typing import List, Callable

from driver_sble_py_klsblelcf import KlSbleLcr

from . import SenseidReader, SenseidReaderDetails
from ..parsers import SenseidTag
from ..parsers.ble import SenseidBleTag

logger = logging.getLogger(__name__)


class SenseidKlSbleLcr(SenseidReader):

    def __init__(self):
        self.details = None
        self.driver = KlSbleLcr()
        self.notification_callback = None
        self.tx_power = 0

    def connect(self, connection_string: str):
        self.driver.connect(connection_string=connection_string)
        self.driver.set_notification_callback(self._sble_notification_callback)
        return True

    def _sble_notification_callback(self, beacon):
        if self.notification_callback is not None:
            self.notification_callback(SenseidBleTag(beacon))
        pass

    def disconnect(self):
        self.driver.disconnect()

    def get_details(self) -> SenseidReaderDetails:
        if self.details is None:
            self.details = SenseidReaderDetails(
                model_name='KL-SBLE-LCR',
                region='EU',
                firmware_version='0.0.1',
                antenna_count=1,
                min_tx_power=10,
                max_tx_power=31.5
            )
            logger.debug(self.details)
        return self.details

    def get_tx_power(self) -> float:
        current_tx_dbm = 10
        logger.debug('get_tx_power: ' + str(current_tx_dbm))
        return current_tx_dbm

    def set_tx_power(self, dbm: float):
        logger.debug('set_tx_power: ' + str(dbm))

    def get_antenna_config(self) -> List[bool]:
        antenna_config_array = [True]
        return antenna_config_array

    def set_antenna_config(self, antenna_config_array: List[bool]):
        logger.debug('set_antenna_config: ' + str(antenna_config_array))

    def start_inventory_async(self, notification_callback: Callable[[SenseidTag], None]):
        self.notification_callback = notification_callback
        return self.driver.configure_cw(True, 500, 500)

    def stop_inventory_async(self):
        return self.driver.configure_cw(False, 0, 0)



import logging
from typing import Callable

from zeroconf import IPVersion, Zeroconf, ServiceBrowser, ServiceStateChange
from .. import SenseidReaderConnectionInfo, SupportedSenseidReader

logger = logging.getLogger(__name__)


class MulticastDnsServiceDiscoveryScanner:

    def __init__(self, notification_callback: Callable[[SenseidReaderConnectionInfo], None], autostart: bool = False):
        self.notification_callback = notification_callback
        self.service_browser: ServiceBrowser | None = None
        self.zeroconf_instance = Zeroconf(ip_version=IPVersion.V4Only)
        self.ips = []

        if autostart:
            self.start()

    def start(self, reset: bool = False):
        if reset:
            self.ips = []

        def on_service_state_change(zeroconf: Zeroconf, service_type: str, name: str, state_change: ServiceStateChange):
            if 'SpeedwayR' in name:
                if state_change is ServiceStateChange.Added:
                    info = zeroconf.get_service_info(service_type, name)
                    if info is not None:
                        ipv4_addresses = info.addresses_by_version(IPVersion.V4Only)
                        if len(ipv4_addresses) > 0:
                            ipv4 = ipv4_addresses[0]
                            if ipv4 not in self.ips:
                                ip_str = str(ipv4[0]) + '.' + str(ipv4[1]) + '.' + str(ipv4[2]) + '.' + str(ipv4[3])
                                logger.info('New Speedway readers found: ' + ip_str)
                                self.ips.append(ip_str)
                                self.notification_callback(
                                    SenseidReaderConnectionInfo(driver=SupportedSenseidReader.OCTANE,
                                                                connection_string=ip_str))
                                self.notification_callback(
                                    SenseidReaderConnectionInfo(driver=SupportedSenseidReader.SPEEDWAY,
                                                                connection_string=ip_str))

            if 'ThingMagic Mercury' in name:
                if state_change is ServiceStateChange.Added:
                    info = zeroconf.get_service_info(service_type, name)
                    if info is not None:
                        ipv4_addresses = info.addresses_by_version(IPVersion.V4Only)
                        if len(ipv4_addresses) > 0:
                            ipv4 = ipv4_addresses[0]
                            if ipv4 not in self.ips:
                                ip_str = str(ipv4[0]) + '.' + str(ipv4[1]) + '.' + str(ipv4[2]) + '.' + str(ipv4[3])
                                logger.info('New Mercury readers found: ' + ip_str)
                                self.ips.append(ip_str)
                                self.notification_callback(
                                    SenseidReaderConnectionInfo(driver=SupportedSenseidReader.SPEEDWAY,
                                                                connection_string=ip_str))

        services = [
            "_http._tcp.local.",
        ]
        self.service_browser = ServiceBrowser(self.zeroconf_instance, services, handlers=[on_service_state_change])

    def stop(self):
        self.service_browser.cancel()
        self.service_browser.join()



import datetime
from dataclasses import dataclass
from enum import Enum
from importlib.resources import files
from typing import List, Dict

import yaml
from dataclasses_json import dataclass_json


class SenseidValueType(Enum):
    UINT8 = 'uint8'
    INT8 = 'int8'
    UINT16 = 'uint16'
    INT16 = 'int16'
    UINT16BE = 'uint16be'
    INT16BE = 'int16be'
    FLOAT = 'float'
    PADDING = 'padding'


class SenseidTransformType(Enum):
    NONE = 'none'
    LINEAR = 'linear'
    THERMISTOR_BETA = 'thermistor-beta'


@dataclass_json
@dataclass
class SenseidBleDataDef:
    magnitude: str
    unit_long: str
    unit_short: str
    type: SenseidValueType
    transform: SenseidTransformType
    coefficients: List[float]


@dataclass_json
@dataclass
class SenseidBleTypeDef:
    name: str
    description: str
    data_def: List[SenseidBleDataDef]
    fw_versions: List[int]


@dataclass_json
@dataclass
class SenseidBleDef:
    version: int
    date: datetime.date
    local_name: str
    types: Dict[int, SenseidBleTypeDef]


# Detect Source or Package mode
top_package = __name__.split('.')[0]
if top_package == 'src':
    senseid_package = files('src.senseid')
else:
    senseid_package = files('senseid')
_senseid_yaml = senseid_package.joinpath('definitions').joinpath('senseid_ble.yaml').read_text()
_senseid_dict = yaml.safe_load(_senseid_yaml)
SENSEID_BLE_DEF: SenseidBleDef = SenseidBleDef.from_dict(_senseid_dict)



import logging
from typing import List, Callable

from nurapi import NUR, NUR_MODULESETUP_FLAGS, NurModuleSetup, NurReaderInfo, NurDeviceCaps, NurTagCount, NurTagData
from nurapi.enums import SETUP_LINK_FREQ, SETUP_RX_DEC, OperationResult

from . import SenseidReader, SenseidReaderDetails
from ..parsers import SenseidTag
from ..parsers.rain import SenseidRainTag

logger = logging.getLogger(__name__)


class SenseidNurapi(SenseidReader):

    def __init__(self):
        self.driver = NUR()
        self.notification_callback = None
        self.device_caps: NurDeviceCaps | None = None
        self.details = None

    def connect(self, connection_string: str):
        self.driver.ConnectSerialPortEx(port_name=connection_string)
        self.driver.set_notification_callback(self._nur_notification_callback)
        self.get_details()

        # Set Senseid compatible mode
        module_setup = NurModuleSetup()
        module_setup.link_freq = SETUP_LINK_FREQ.BLF_256
        module_setup.rx_decoding = SETUP_RX_DEC.MILLER_4
        self.driver.SetModuleSetup(setup_flags=[NUR_MODULESETUP_FLAGS.NUR_SETUP_LINKFREQ,
                                               NUR_MODULESETUP_FLAGS.NUR_SETUP_RXDEC], module_setup=module_setup)

        # Set MAX TX Power
        self.set_tx_power(self.details.max_tx_power)

        # Enable first antenna
        antenna_config = [False] * self.details.antenna_count
        antenna_config[0] = True
        self.set_antenna_config(antenna_config_array=antenna_config)
        return True

    def _nur_notification_callback(self, inventory_stream_data):
        # If stream stopped, restart
        if inventory_stream_data.stopped:
            self.driver.StartInventoryStream(rounds=10, q=0, session=0)

        # Check number of tags read
        tag_count = self.driver.GetTagCount()
        # Get data of read tags
        for idx in range(tag_count):
            tag_data = self.driver.GetTagData(idx=idx)
            if self.notification_callback is not None:
                self.notification_callback(SenseidRainTag(epc=tag_data.epc))
        self.driver.ClearTags()

    def disconnect(self):
        self.driver.Disconnect()

    def get_details(self) -> SenseidReaderDetails:
        if self.details is None:
            reader_info = self.driver.GetReaderInfo()
            self.device_caps = self.driver.GetDeviceCaps()

            module_setup = self.driver.GetModuleSetup(setup_flags=[NUR_MODULESETUP_FLAGS.NUR_SETUP_REGION])

            self.details = SenseidReaderDetails(
                model_name=reader_info.name,
                region=module_setup.region_id.name,
                firmware_version=str(reader_info.sw_ver_major) + '.' + str(reader_info.sw_ver_minor),
                antenna_count=reader_info.num_antennas,
                min_tx_power=self.device_caps.maxTxdBm - (self.device_caps.txSteps-1) * self.device_caps.txAttnStep,
                max_tx_power=self.device_caps.maxTxdBm
            )
            logger.debug(self.details)
        return self.details

    def get_tx_power(self) -> float:
        # Only supporting same power on all antennas
        module_setup = self.driver.GetModuleSetup(setup_flags=[NUR_MODULESETUP_FLAGS.NUR_SETUP_TXLEVEL])
        current_tx_dbm = self.device_caps.maxTxdBm - module_setup.tx_level / self.device_caps.txAttnStep
        logger.debug('get_tx_power: ' + str(current_tx_dbm))
        return current_tx_dbm

    def set_tx_power(self, dbm: float):
        logger.debug('set_tx_power: ' + str(dbm))
        # Only supporting same power on all antennas
        if self.details is None:
            self.get_details()
        if dbm > self.details.max_tx_power:
            dbm = self.details.max_tx_power
            logger.warning('Power set to max power: ' + str(dbm))
        if dbm < self.details.min_tx_power:
            dbm = self.details.min_tx_power
            logger.warning('Power set to min power: ' + str(dbm))

        module_setup = NurModuleSetup()
        module_setup.tx_level = int((self.device_caps.maxTxdBm - dbm) * self.device_caps.txAttnStep)
        self.driver.SetModuleSetup(setup_flags=[NUR_MODULESETUP_FLAGS.NUR_SETUP_TXLEVEL], module_setup=module_setup)

    def get_antenna_config(self) -> List[bool]:
        module_setup = self.driver.GetModuleSetup(setup_flags=[NUR_MODULESETUP_FLAGS.NUR_SETUP_ANTMASKEX])
        antenna_mask = module_setup.antenna_mask_ex
        antenna_config_array = []
        for i in range(self.details.antenna_count):
            antenna_bit = (antenna_mask >> i) & 0b1
            antenna_config_array.append(bool(antenna_bit))
        logger.debug('get_antenna_config: ' + str(antenna_config_array))
        return antenna_config_array

    def set_antenna_config(self, antenna_config_array: List[bool]):
        logger.debug('set_antenna_config: ' + str(antenna_config_array))
        if not (True in antenna_config_array):
            antenna_config_array[0] = True
            logger.warning('At least one antenna needs to be active. Enabling antenna 1.')
        antenna_mask = 0
        for idx, antenna_config in enumerate(antenna_config_array):
            antenna_mask |= antenna_config << idx
        module_setup = NurModuleSetup()
        module_setup.antenna_mask_ex = antenna_mask
        module_setup.selected_antenna = -1  # Automatic selection
        self.driver.SetModuleSetup(setup_flags=[NUR_MODULESETUP_FLAGS.NUR_SETUP_ANTMASKEX,
                                               NUR_MODULESETUP_FLAGS.NUR_SETUP_SELECTEDANT],
                                   module_setup=module_setup)

    def start_inventory_async(self, notification_callback: Callable[[SenseidTag], None]):
        self.notification_callback = notification_callback
        return self.driver.StartInventoryStream(rounds=10, q=0, session=0)

    def stop_inventory_async(self):
        return self.driver.StopInventoryStream()
